import datetime
from typing import Optional

from gumo.task.domain import GumoTask
from gumo.task_emulator.domain import GumoTaskProcess
from gumo.task_emulator.domain import ProcessHistory


class TaskJSONEncoder:
    def __init__(
            self,
            task: GumoTask,
    ):
        self._task = task

    def datetime_to_json(self, t: datetime.datetime) -> Optional[str]:
        if t is None:
            return
        return t.isoformat()

    def to_json(self) -> dict:
        j = {
            'key': self._task.key.key_path(),
            'keyLiteral': self._task.key.key_literal(),
            'relativeURI': self._task.relative_uri,
            'method': self._task.method,
            'payload': self._task.payload,
            'scheduleTime': self.datetime_to_json(self._task.schedule_time),
            'createdAt': self.datetime_to_json(self._task.created_at),
            'queueName': self._task.queue_name,
        }
        return j


class TaskProcessJSONEncoder:
    def __init__(
            self,
            task_process: GumoTaskProcess,
    ):
        self._task_process = task_process

    def datetime_to_json(self, t: datetime.datetime) -> Optional[str]:
        if t is None:
            return
        return t.isoformat()

    def to_json(self) -> dict:
        j = {
            'key': self._task_process.key.key_path(),
            'keyLiteral': self._task_process.key.key_literal(),
            'relativeURI': self._task_process.relative_uri,
            'method': self._task_process.method,
            'payload': self._task_process.payload,
            'scheduleTime': self.datetime_to_json(self._task_process.schedule_time),
            'createdAt': self.datetime_to_json(self._task_process.created_at),
            'updatedAt': self.datetime_to_json(self._task_process.updated_at),
            'queueName': self._task_process.queue_name,
            'state': self._task_process.state.value,
            'attempts': self._task_process.attempts,
            'lastRunAt': self.datetime_to_json(self._task_process.last_run_at),
            'runAt': self.datetime_to_json(self._task_process.run_at),
            'lockedAt': self.datetime_to_json(self._task_process.locked_at),
            'succeededAt': self.datetime_to_json(self._task_process.succeeded_at),
            'failedAt': self.datetime_to_json(self._task_process.failed_at),
            'histories': [
                self._history_to_json(history=history) for history in self._task_process.histories
            ],
        }
        return j

    def _history_to_json(self, history: ProcessHistory) -> dict:
        j = {
            'startedAt': self.datetime_to_json(history.started_at),
            'durationSeconds': history.duration_seconds,
            'statusCode': history.status_code,
            'requestHeaders': history.request_headers,
            'requestBody': history.request_body,
            'responseHeaders': history.response_headers,
            'responseBody': history.response_body,
            'errorMessage': history.error_message,
        }
        return j



from logging import getLogger
import flask.views

from gumo.core.injector import injector

from gumo.datastore import EntityKeyFactory

from gumo.task_emulator.application import TaskExecuteService
from gumo.task_emulator.application import TaskMarkAsFailedService
from gumo.task_emulator.application.task import TaskProcessBulkCreateService
from gumo.task_emulator.application.task.encoder import TaskJSONEncoder
from gumo.task_emulator.application.task.encoder import TaskProcessJSONEncoder
from gumo.task_emulator.application.task.repository import TaskRepository
from gumo.task_emulator.application.task.repository import TaskProcessRepository
from gumo.task_emulator.application.task.repository import TaskProcessSummaryRepository
from gumo.task_emulator.domain import TaskState

logger = getLogger(__name__)
emulator_api_blueprint = flask.Blueprint('task-emulator', __name__)


class TasksView(flask.views.MethodView):
    _repository  = injector.get(TaskRepository)  # type: TaskRepository

    def get(self):
        tasks = self._repository.fetch_tasks(limit=100)

        return flask.jsonify({
            'results': [TaskJSONEncoder(task).to_json() for task in tasks]
        })

    def post(self):
        return 'ok'


class TaskDetailView(flask.views.MethodView):
    _repository = injector.get(TaskProcessRepository)  # type: TaskProcessRepository

    def get(self, key):
        task = self._repository.fetch_by_key(
            key=EntityKeyFactory().build_from_key_path(key)
        )

        return flask.jsonify(TaskProcessJSONEncoder(task).to_json())


class TasksEmulatorEnqueue(flask.views.MethodView):
    _task_process_create_service = injector.get(TaskProcessBulkCreateService)  # type: TaskProcessBulkCreateService

    def get(self):
        result = self._task_process_create_service.execute()
        return flask.jsonify(result)

    def post(self):
        result = self._task_process_create_service.execute()
        return flask.jsonify(result)


class TaskProcessesView(flask.views.MethodView):
    _repository = injector.get(TaskProcessRepository)  # type: TaskProcessRepository

    def get(self):
        task_processes = self._repository.fetch_tasks()

        return flask.jsonify({
            'tasks': [
                TaskProcessJSONEncoder(task_process=task_process).to_json()
                for task_process in task_processes
            ]
        })


class TaskSummaryView(flask.views.MethodView):
    _repository = injector.get(TaskProcessSummaryRepository)  # type: TaskProcessSummaryRepository

    def get(self):
        summary = self._repository.fetch_summary()

        return flask.jsonify({
            'summary': summary.to_dict(),
        })


class QueuedTasksView(flask.views.MethodView):
    _repository = injector.get(TaskProcessRepository)  # type: TaskProcessRepository

    def get(self):
        task_processes = self._repository.fetch_tasks_by_state(state=TaskState.QUEUED)

        return flask.jsonify({
            'tasks': [
                TaskProcessJSONEncoder(task_process).to_json()
                for task_process in task_processes
            ]
        })


class ExecuteTaskView(flask.views.MethodView):
    _service = injector.get(TaskExecuteService)  # type: TaskExecuteService

    def get(self, key):
        result = self._service.execute(key=EntityKeyFactory().build_from_key_path(key_path=key))
        return flask.jsonify(result.to_json())


class TaskRemoveView(flask.views.MethodView):
    _service = injector.get(TaskMarkAsFailedService)  # type: TaskMarkAsFailedService

    def get(self, key):
        task = self._service.execute(
            key=EntityKeyFactory().build_from_key_path(key)
        )

        return flask.jsonify(TaskProcessJSONEncoder(task).to_json())


class TaskCleanupView(flask.views.MethodView):
    _repository = injector.get(TaskProcessRepository)  # type: TaskProcessRepository

    def get(self):
        self._repository.cleanup_finished_tasks()

        return flask.jsonify({'cleanup': 'ok'})


emulator_api_blueprint.add_url_rule(
    '/api/tasks',
    view_func=TasksView.as_view(name='tasks'),
    methods=['GET', 'POST']
)

emulator_api_blueprint.add_url_rule(
    '/api/tasks/<key>',
    view_func=TaskDetailView.as_view(name='tasks/detail'),
    methods=['GET']
)

emulator_api_blueprint.add_url_rule(
    '/api/task_emulator/enqueue',
    view_func=TasksEmulatorEnqueue.as_view(name='task_emulator/enqueue'),
    methods=['GET', 'POST']
)

emulator_api_blueprint.add_url_rule(
    '/api/task_emulator/tasks',
    view_func=TaskProcessesView.as_view(name='task_emulator/tasks'),
    methods=['GET']
)

emulator_api_blueprint.add_url_rule(
    '/api/task_emulator/tasks/summary',
    view_func=TaskSummaryView.as_view(name='task_emulator/tasks/summary'),
    methods=['GET']
)

emulator_api_blueprint.add_url_rule(
    '/api/task_emulator/tasks/queued',
    view_func=QueuedTasksView.as_view(name='task_emulator/tasks/queued'),
    methods=['GET']
)

emulator_api_blueprint.add_url_rule(
    '/api/task_emulator/tasks/<key>/execute',
    view_func=ExecuteTaskView.as_view(name='task_emulator/tasks/execute'),
    methods=['GET']
)

emulator_api_blueprint.add_url_rule(
    '/api/task_emulator/tasks/<key>/remove',
    view_func=TaskRemoveView.as_view(name='task_emulator/tasks/remove'),
    methods=['GET']
)

emulator_api_blueprint.add_url_rule(
    '/api/task_emulator/cleanup',
    view_func=TaskCleanupView.as_view(name='task_emulator/cleanup'),
    methods=['GET']
)



import os
import urllib.request
import urllib.error
import json
import datetime
from logging import getLogger
from injector import inject

from gumo.task_emulator.application import TaskExecuteRunner
from gumo.task_emulator.domain.configuration import TaskEmulatorConfiguration
from gumo.task_emulator.domain import GumoTaskProcess
from gumo.task_emulator.domain import ProcessRequest
from gumo.task_emulator.domain import ProcessHistory

logger = getLogger(__name__)


class TaskExecuteRunnerImpl(TaskExecuteRunner):
    @inject
    def __init__(
            self,
            task_emulator_configuration: TaskEmulatorConfiguration,
    ):
        self._task_emulator_configuration = task_emulator_configuration

    def _build_request_headers(self, task_process: GumoTaskProcess) -> dict:
        headers = {
            'Content-Type': 'application/json',
            'X-AppEngine-QueueName': '',
            'X-AppEngine-TaskName': task_process.key.name(),
            'X-AppEngine-TaskRetryCount': str(task_process.attempts),
            'X-AppEngine-TaskExecutionCount': str(task_process.attempts + 1),
            'X-AppEngine-TaskETA': str(int(task_process.run_at.timestamp())),
        }

        if len(task_process.histories) > 0:
            previous_history = task_process.histories[-1]
            headers['X-AppEngine-TaskPreviousResponse'] = str(previous_history.status_code)

        return headers

    def _build_request(self, task_process: GumoTaskProcess) -> ProcessRequest:
        target_url = 'http://{host}:{port}{relative_uri}'.format(
            host=self._task_emulator_configuration.server_host,
            port=self._task_emulator_configuration.server_port,
            relative_uri=task_process.relative_uri,
        )
        request_headers = self._build_request_headers(task_process=task_process)
        request_body = json.dumps(task_process.payload)

        return ProcessRequest(
            method=task_process.method,
            url=target_url,
            headers=request_headers,
            data=request_body,
        )

    def _run_request(self, process_request: ProcessRequest) -> ProcessHistory:
        request = urllib.request.Request(
            url=process_request.url,
            method=process_request.method,
            data=process_request.data_as_bytes(),
            headers=process_request.headers,
        )

        started_at = datetime.datetime.utcnow()

        status_code = None
        response_headers = None
        response_body = None
        error_message = None

        try:
            with urllib.request.urlopen(request) as response:
                status_code = response.status
                response_body = response.read().decode('utf-8')
        except urllib.error.HTTPError as e:
            status_code = e.code
            error_message = e
        except urllib.error.URLError as e:
            error_message = e
        except Exception as e:
            error_message = e

        finished_at = datetime.datetime.utcnow()

        return ProcessHistory(
            started_at=started_at,
            duration_seconds=(finished_at - started_at).total_seconds(),
            method=process_request.method,
            url=process_request.url,
            data=process_request.data,
            status_code=status_code,
            request_headers=process_request.headers,
            request_body=process_request.data,
            response_headers=response_headers,
            response_body=response_body,
            error_message=str(error_message) if error_message else None,
        )

    def run(self, task_process: GumoTaskProcess) -> ProcessHistory:
        request = self._build_request(task_process=task_process)
        history = self._run_request(process_request=request)

        return history



from typing import List
from typing import Optional

from gumo.datastore import EntityKey
from gumo.task.domain import GumoTask
from gumo.task_emulator.domain import GumoTaskProcess
from gumo.task_emulator.domain import TaskState
from gumo.task_emulator.domain import QueueStatusCollection


class TaskRepository:
    def fetch_tasks(self, limit: int = 10) -> List[GumoTask]:
        raise NotImplementedError()

    def save(self, task: GumoTask) -> GumoTask:
        raise NotImplementedError()

    def delete(self, key: EntityKey):
        raise NotImplementedError()


class TaskProcessRepository:
    def fetch_by_key(self, key: EntityKey) -> GumoTaskProcess:
        raise NotImplementedError()

    def fetch_tasks_by_state(self, state: TaskState, limit: Optional[int] = None) -> List[GumoTaskProcess]:
        raise NotImplementedError()

    def fetch_tasks(
            self,
            queue_name: Optional[str] = None,
            limit: Optional[int] = None,
    ) -> List[GumoTaskProcess]:
        raise NotImplementedError()

    def save(self, task_process: GumoTaskProcess):
        raise NotImplementedError()

    def cleanup_finished_tasks(self):
        raise NotImplementedError()


class TaskProcessSummaryRepository:
    def fetch_summary(self) -> QueueStatusCollection:
        raise NotImplementedError()



from logging import getLogger

import os
import flask.views
import flasgger

from gumo.task_emulator._configuration import configure
from gumo.task_emulator.domain.configuration import TaskEmulatorConfiguration
from gumo.task_emulator.presentation.cli.worker import BackgroundWorker


def task_emulator_app():
    from gumo.task_emulator.presentation.restapi import emulator_api_blueprint
    from gumo.task_emulator.presentation.ui import emulator_ui_blueprint

    flask_app = flask.Flask(__name__)
    flask_app.config['JSON_AS_ASCII'] = False
    flask_app.config['SWAGGER'] = {
        'title': 'Task Emulator',
        'doc_dir': os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            'apidoc',
        )
    }
    flask_app.register_blueprint(emulator_api_blueprint)
    flask_app.register_blueprint(emulator_ui_blueprint)

    flasgger.Swagger(
        flask_app,
        template_file=os.path.join(flask_app.config['SWAGGER']['doc_dir'], 'template.yml')
    )

    return flask_app, BackgroundWorker.get_instance()


__all__ = [
    configure.__name__,

    TaskEmulatorConfiguration.__name__,

    task_emulator_app.__name__,
]

logger = getLogger('gumo.task_emulator')



import setuptools


name = 'gumo-task-emulator'
version = '0.0.3'
description = 'Gumo Task Emulator Library'
dependencies = [
    f'gumo-core >= {version}',
    f'gumo-datastore >= {version}',
    f'gumo-task >= {version}',
    'Flask >= 1.0.2',
    'flasgger >= 0.9.1',
]

with open("README.md", "r") as fh:
    long_description = fh.read()

packages = [
    package for package in setuptools.find_packages()
    if package.startswith('gumo')
]

namespaces = ['gumo', 'gumo.task_emulator']

setuptools.setup(
    name=name,
    version=version,
    author="Gumo Project Team",
    author_email="gumo-organizer@levii.co.jp",
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gumo-py/gumo",
    packages=packages,
    namespaces=namespaces,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=dependencies,
)



from logging import getLogger
from typing import Union

from gumo.core.injector import injector
from gumo.task_emulator.domain.configuration import TaskEmulatorConfiguration
from gumo.task_emulator.bind import task_emulator_bind


logger = getLogger('gumo.task')


class ConfigurationFactory:
    DEFAULT_SERVER_PORT = 8080

    @classmethod
    def build(
            cls,
            server_host: str,
            server_port: Union[str, int, None],
    ) -> TaskEmulatorConfiguration:
        if isinstance(server_port, str):
            if server_port.isdigit():
                port = int(server_port)
            else:
                raise ValueError(f'Invalid format of server_port={server_port}')
        elif isinstance(server_port, int):
            port = server_port
        else:
            port = cls.DEFAULT_SERVER_PORT

        return TaskEmulatorConfiguration(
            server_host=server_host,
            server_port=port,
        )


def configure(
        server_host: str,
        server_port: Union[str, int, None] = None,
) -> TaskEmulatorConfiguration:
    config = ConfigurationFactory.build(
        server_host=server_host,
        server_port=server_port,
    )
    logger.debug(f'Gumo.TaskEmulator is configured, config={config}')

    injector.binder.bind(TaskEmulatorConfiguration, to=config)
    injector.binder.install(task_emulator_bind)

    return config



from gumo.task_emulator.application.task.repository import TaskRepository
from gumo.task_emulator.infrastructure.repository import DatastoreTaskRepository

from gumo.task_emulator.application.task.repository import TaskProcessRepository
from gumo.task_emulator.infrastructure.repository import DatastoreTaskProcessRepository

from gumo.task_emulator.application import TaskExecuteRunner
from gumo.task_emulator.infrastructure.execute_service import TaskExecuteRunnerImpl

from gumo.task_emulator.application.task.repository import TaskProcessSummaryRepository
from gumo.task_emulator.infrastructure.repository import DatastoreTaskProcessSummaryRepository


def task_emulator_bind(binder):
    binder.bind(TaskRepository, to=DatastoreTaskRepository)
    binder.bind(TaskProcessRepository, to=DatastoreTaskProcessRepository)
    binder.bind(TaskExecuteRunner, to=TaskExecuteRunnerImpl)
    binder.bind(TaskProcessSummaryRepository, to=DatastoreTaskProcessSummaryRepository)
